package io.tswf.groovy.bettergroovy.asm.expansion.generator.conflict

import groovy.transform.CompileStatic
import io.tswf.groovy.bettergroovy.asm.expansion.exception.ConflictResolveException
import io.tswf.groovy.bettergroovy.asm.expansion.exception.IConflictResolveExceptionBypassMarker
import io.tswf.groovy.bettergroovy.asm.expansion.meta.ExpansionConflictResolveStrategy
import io.tswf.groovy.bettergroovy.asm.expansion.meta.ExpansionMethodMetadata
import org.objectweb.asm.tree.ClassNode
import org.objectweb.asm.tree.MethodNode

@CompileStatic
class ConflictResolver {

    @Lazy
    static Map<ExpansionConflictResolveStrategy, IConflictResolveStrategyLogic> conflictResolvers = ServiceLoader.load(IConflictResolveStrategyLogic)
        .toMapBy { it.enumValue.ensureNotNull { new NoSuchElementException("Can't find enum value for ${it}") } }
        .tap { resolvers ->
            def missingKeys = ExpansionConflictResolveStrategy.values()
                .stream()
                    .filter { !resolvers.containsKey(it) }
                    .toSet()

            if (missingKeys.isNotNullOrEmpty()) {
                throw new RuntimeException("There is no conflict resolve logic found for strategies: ${missingKeys}")
            }
        }

    static boolean resolveExpansionConflict(MethodNode conflictingMethod, ClassNode targetClass, ExpansionMethodMetadata expansionMethodMeta) {
        try {
            def conflictResolver = conflictResolvers[expansionMethodMeta.conflictResolveStrategy]
                    .ensureNotNull { new RuntimeException("There is no conflict resolve strategy logic found for: ${expansionMethodMeta.conflictResolveStrategy.name()}!") }

            def conflictResolveContext = new ConflictResolveContext(
                    expansionMethodMeta: expansionMethodMeta,
                    targetClass: targetClass,
                    conflictingMethod: conflictingMethod
            )

            conflictResolver.handleConflictResolve(conflictResolveContext)

            return conflictResolveContext.canContinue

        } catch (Throwable e) {
            // Если выброшенное исключение запланировано, то его обрабатывать кастомно не надо
            if (e instanceof IConflictResolveExceptionBypassMarker) {
                throw e
            }

            throw new ConflictResolveException(
                    "Exception occurred while trying to resolve conflict for ${conflictingMethod.javaFullName} in ${targetClass.javaName} and ${expansionMethodMeta.expansion.expansionClass.javaName}!",
                    e
            )
        }
    }

}
